# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #
from enum import Enum


class OscilloscopeDataSource(Enum):
    """
    Kind of channel to record in the Oscilloscope.
    """

    SETTING = 0
    IO = 1



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from typing import TYPE_CHECKING
from ..call import call, call_async, call_sync

from ..protobufs import main_pb2

if TYPE_CHECKING:
    from .device import Device


class AllAxes:
    """
    Represents all axes of motion associated with a device.
    """

    @property
    def device(self) -> 'Device':
        """
        Device that controls this axis.
        """
        return self._device

    def __init__(self, device: 'Device'):
        self._device = device

    def home(
            self,
            wait_until_idle: bool = True
    ) -> None:
        """
        Homes all axes. Axes return to their homing positions.

        Args:
            wait_until_idle: Determines whether function should return after the movement is finished or just started.
        """
        request = main_pb2.DeviceHomeRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.axis = 0
        request.wait_until_idle = wait_until_idle
        call("device/home", request)

    async def home_async(
            self,
            wait_until_idle: bool = True
    ) -> None:
        """
        Homes all axes. Axes return to their homing positions.

        Args:
            wait_until_idle: Determines whether function should return after the movement is finished or just started.
        """
        request = main_pb2.DeviceHomeRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.axis = 0
        request.wait_until_idle = wait_until_idle
        await call_async("device/home", request)

    def stop(
            self,
            wait_until_idle: bool = True
    ) -> None:
        """
        Stops ongoing axes movement. Decelerates until zero speed.

        Args:
            wait_until_idle: Determines whether function should return after the movement is finished or just started.
        """
        request = main_pb2.DeviceStopRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.axis = 0
        request.wait_until_idle = wait_until_idle
        call("device/stop", request)

    async def stop_async(
            self,
            wait_until_idle: bool = True
    ) -> None:
        """
        Stops ongoing axes movement. Decelerates until zero speed.

        Args:
            wait_until_idle: Determines whether function should return after the movement is finished or just started.
        """
        request = main_pb2.DeviceStopRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.axis = 0
        request.wait_until_idle = wait_until_idle
        await call_async("device/stop", request)

    def wait_until_idle(
            self,
            throw_error_on_fault: bool = True
    ) -> None:
        """
        Waits until all axes of device stop moving.

        Args:
            throw_error_on_fault: Determines whether to throw error when fault is observed.
        """
        request = main_pb2.DeviceWaitUntilIdleRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.axis = 0
        request.throw_error_on_fault = throw_error_on_fault
        call("device/wait_until_idle", request)

    async def wait_until_idle_async(
            self,
            throw_error_on_fault: bool = True
    ) -> None:
        """
        Waits until all axes of device stop moving.

        Args:
            throw_error_on_fault: Determines whether to throw error when fault is observed.
        """
        request = main_pb2.DeviceWaitUntilIdleRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.axis = 0
        request.throw_error_on_fault = throw_error_on_fault
        await call_async("device/wait_until_idle", request)

    def park(
            self
    ) -> None:
        """
        Parks the device in anticipation of turning the power off.
        It can later be powered on, unparked, and moved without first having to home it.
        """
        request = main_pb2.AxisEmptyRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.axis = 0
        call("device/park", request)

    async def park_async(
            self
    ) -> None:
        """
        Parks the device in anticipation of turning the power off.
        It can later be powered on, unparked, and moved without first having to home it.
        """
        request = main_pb2.AxisEmptyRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.axis = 0
        await call_async("device/park", request)

    def unpark(
            self
    ) -> None:
        """
        Unparks the device. The device will now be able to move.
        """
        request = main_pb2.AxisEmptyRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.axis = 0
        call("device/unpark", request)

    async def unpark_async(
            self
    ) -> None:
        """
        Unparks the device. The device will now be able to move.
        """
        request = main_pb2.AxisEmptyRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.axis = 0
        await call_async("device/unpark", request)

    def is_busy(
            self
    ) -> bool:
        """
        Returns bool indicating whether any axis is executing a motion command.

        Returns:
            True if any axis is currently executing a motion command.
        """
        request = main_pb2.AxisEmptyRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.axis = 0
        response = main_pb2.BoolResponse()
        call("device/is_busy", request, response)
        return response.value

    async def is_busy_async(
            self
    ) -> bool:
        """
        Returns bool indicating whether any axis is executing a motion command.

        Returns:
            True if any axis is currently executing a motion command.
        """
        request = main_pb2.AxisEmptyRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.axis = 0
        response = main_pb2.BoolResponse()
        await call_async("device/is_busy", request, response)
        return response.value

    def is_homed(
            self
    ) -> bool:
        """
        Returns bool indicating whether all axes have position reference and were homed.

        Returns:
            True if all axes have position reference and were homed.
        """
        request = main_pb2.AxisEmptyRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.axis = 0
        response = main_pb2.BoolResponse()
        call("device/is_homed", request, response)
        return response.value

    async def is_homed_async(
            self
    ) -> bool:
        """
        Returns bool indicating whether all axes have position reference and were homed.

        Returns:
            True if all axes have position reference and were homed.
        """
        request = main_pb2.AxisEmptyRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.axis = 0
        response = main_pb2.BoolResponse()
        await call_async("device/is_homed", request, response)
        return response.value

    def __repr__(
            self
    ) -> str:
        """
        Returns a string that represents the axes.

        Returns:
            A string that represents the axes.
        """
        request = main_pb2.AxisEmptyRequest()
        request.interface_id = self.device.connection.interface_id
        request.device = self.device.device_address
        request.axis = 0
        response = main_pb2.StringResponse()
        call_sync("device/all_axes_to_string", request, response)
        return response.value



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from typing import Union
from .set_peripheral_state_exception_data import SetPeripheralStateExceptionData
from ..protobufs import main_pb2
from .motion_lib_exception import MotionLibException


class SetPeripheralStateFailedException(MotionLibException):
    """
    Thrown when an axis cannot be set to the supplied state.
    """

    @property
    def details(self) -> SetPeripheralStateExceptionData:
        """
        Additional data for SetPeripheralStateFailedException
        """
        return self._details

    def __init__(self, message: str, custom_data: Union[bytes, SetPeripheralStateExceptionData]):
        MotionLibException.__init__(self, message)

        if isinstance(custom_data, SetPeripheralStateExceptionData):
            self._details = custom_data
        else:
            protobuf_obj = main_pb2.SetPeripheralStateExceptionData()
            protobuf_obj.ParseFromString(custom_data)
            self._details = SetPeripheralStateExceptionData.from_protobuf(protobuf_obj)



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #
# pylint: disable=W0201

from typing import Optional  # pylint: disable=unused-import
from ..protobufs import main_pb2


class AxisDefinition:
    """
    Defines an axis of the translator.
    """

    def __init__(
            self: 'AxisDefinition',
            peripheral_id: int,
            microstep_resolution: Optional[int] = None
    ) -> None:
        self._peripheral_id = peripheral_id
        self._microstep_resolution = microstep_resolution

    @property
    def peripheral_id(self) -> int:
        """
        ID of the peripheral.
        """

        return self._peripheral_id

    @peripheral_id.setter
    def peripheral_id(self, value: int) -> None:
        self._peripheral_id = value

    @property
    def microstep_resolution(self) -> Optional[int]:
        """
        Microstep resolution of the axis.
        Can be obtained by reading the resolution setting.
        Leave empty if the axis does not have the setting.
        """

        return self._microstep_resolution

    @microstep_resolution.setter
    def microstep_resolution(self, value: Optional[int]) -> None:
        self._microstep_resolution = value

    def __repr__(self) -> str:
        return str(self.__dict__)

    @staticmethod
    def from_protobuf(
        pb_data: main_pb2.TranslatorAxisDefinition
    ) -> 'AxisDefinition':
        instance = AxisDefinition.__new__(
            AxisDefinition
        )  # type: AxisDefinition
        instance.peripheral_id = pb_data.peripheral_id
        instance.microstep_resolution = pb_data.microstep_resolution
        return instance

    @staticmethod
    def to_protobuf(source: 'Optional[AxisDefinition]') -> main_pb2.TranslatorAxisDefinition:
        pb_data = main_pb2.TranslatorAxisDefinition()

        if source is None:
            return pb_data

        if not isinstance(source, AxisDefinition):
            raise TypeError("Provided value is not AxisDefinition.")

        pb_data.peripheral_id = source.peripheral_id
        if source.microstep_resolution is not None:
            pb_data.microstep_resolution = source.microstep_resolution
        return pb_data



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #
# pylint: disable=W0201

from typing import Optional  # pylint: disable=unused-import
from ..protobufs import main_pb2
from .process_controller_source_sensor import ProcessControllerSourceSensor


class ProcessControllerSource:
    """
    The source used by a process in a closed-loop mode.
    """

    def __init__(
            self: 'ProcessControllerSource',
            sensor: ProcessControllerSourceSensor,
            port: int
    ) -> None:
        self._sensor = sensor
        self._port = port

    @property
    def sensor(self) -> ProcessControllerSourceSensor:
        """
        The type of input sensor.
        """

        return self._sensor

    @sensor.setter
    def sensor(self, value: ProcessControllerSourceSensor) -> None:
        self._sensor = value

    @property
    def port(self) -> int:
        """
        The specific input to use.
        """

        return self._port

    @port.setter
    def port(self, value: int) -> None:
        self._port = value

    def __repr__(self) -> str:
        return str(self.__dict__)

    @staticmethod
    def from_protobuf(
        pb_data: main_pb2.ProcessControllerSource
    ) -> 'ProcessControllerSource':
        instance = ProcessControllerSource.__new__(
            ProcessControllerSource
        )  # type: ProcessControllerSource
        instance.sensor = ProcessControllerSourceSensor(pb_data.sensor)
        instance.port = pb_data.port
        return instance

    @staticmethod
    def to_protobuf(source: 'Optional[ProcessControllerSource]') -> main_pb2.ProcessControllerSource:
        pb_data = main_pb2.ProcessControllerSource()

        if source is None:
            return pb_data

        if not isinstance(source, ProcessControllerSource):
            raise TypeError("Provided value is not ProcessControllerSource.")

        pb_data.sensor = source.sensor.value if source.sensor is not None else 0
        pb_data.port = source.port
        return pb_data



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #
# pylint: disable=W0201

from typing import List  # pylint: disable=unused-import
from ..protobufs import main_pb2


class SetPeripheralStateExceptionData:
    """
    Contains additional data for a SetPeripheralStateFailedException.
    """

    @property
    def axis_number(self) -> int:
        """
        The number of axis where the exception originated.
        """

        return self._axis_number

    @axis_number.setter
    def axis_number(self, value: int) -> None:
        self._axis_number = value

    @property
    def settings(self) -> List[str]:
        """
        A list of settings which could not be set.
        """

        return self._settings

    @settings.setter
    def settings(self, value: List[str]) -> None:
        self._settings = value

    @property
    def servo_tuning(self) -> str:
        """
        The reason servo tuning could not be set.
        """

        return self._servo_tuning

    @servo_tuning.setter
    def servo_tuning(self, value: str) -> None:
        self._servo_tuning = value

    @property
    def stored_positions(self) -> List[str]:
        """
        The reasons stored positions could not be set.
        """

        return self._stored_positions

    @stored_positions.setter
    def stored_positions(self, value: List[str]) -> None:
        self._stored_positions = value

    @property
    def storage(self) -> List[str]:
        """
        The reasons storage could not be set.
        """

        return self._storage

    @storage.setter
    def storage(self, value: List[str]) -> None:
        self._storage = value

    def __repr__(self) -> str:
        return str(self.__dict__)

    @staticmethod
    def from_protobuf(
        pb_data: main_pb2.SetPeripheralStateExceptionData
    ) -> 'SetPeripheralStateExceptionData':
        instance = SetPeripheralStateExceptionData.__new__(
            SetPeripheralStateExceptionData
        )  # type: SetPeripheralStateExceptionData
        instance.axis_number = pb_data.axis_number
        instance.settings = list(pb_data.settings)
        instance.servo_tuning = pb_data.servo_tuning
        instance.stored_positions = list(pb_data.stored_positions)
        instance.storage = list(pb_data.storage)
        return instance



import pytest
import asyncio
from concurrent.futures import Future
from rx import operators as rxop

from zaber_motion.call import call, call_sync, call_async
from zaber_motion.events import events
from zaber_motion.protobufs import main_pb2
from zaber_motion import RequestTimeoutException, InvalidPacketException


def test_request_response():
    request = main_pb2.TestRequest()
    request.data_ping = "Hello"

    response = main_pb2.TestResponse()
    call("test/request", request, response)

    assert response.data_pong == 'Hello...'

@pytest.mark.asyncio
async def test_request_response_async():
    request = main_pb2.TestRequest()
    request.data_ping = "Hello"

    response = main_pb2.TestResponse()
    await call_async("test/request", request, response)

    assert response.data_pong == 'Hello...'

@pytest.mark.asyncio
async def test_request_response_async_cancel():
    """
    Tests that cancellation does not cause an issue for async call
    """
    async def task_func():
        request = main_pb2.TestRequest()
        request.data_ping = "Hello"

        response = main_pb2.TestResponse()

        asyncio.current_task().cancel()
        await call_async("test/request", request, response)

    try:
        task = asyncio.create_task(task_func())
    except AttributeError:
        pytest.skip('Unsupported Python version')

    try:
        await task
    except asyncio.CancelledError:
        pass

def test_request_response_sync():
    request = main_pb2.TestRequest()
    request.data_ping = "Hello"

    response = main_pb2.TestResponse()
    call_sync("test/request", request, response)

    assert response.data_pong == 'Hello...'

def test_request_error():
    request = main_pb2.TestRequest()
    request.data_ping = "Hello"
    request.return_error = True

    try:
        response = main_pb2.TestResponse()
        call("test/request", request, response)
        assert False, "No error thrown"
    except RequestTimeoutException as e:
        assert e.message == "Device has not responded in given timeout"
        assert str(e) == "RequestTimeoutException: Device has not responded in given timeout"

def test_request_error_with_data():
    request = main_pb2.TestRequest()
    request.return_error_with_data = True

    try:
        response = main_pb2.TestResponse()
        call("test/request", request, response)
        assert False, "No error thrown"
    except InvalidPacketException as e:
        expected_packet = '123'
        expected_reason = 'For test'
        expected_message = 'Cannot parse incoming packet: "{}" ({})'.format(expected_packet, expected_reason)
        assert e.message == expected_message
        assert str(e) == 'InvalidPacketException: ' + expected_message
        assert e.details.packet == expected_packet
        assert e.details.reason == expected_reason

def test_request_no_response():
    response = call("test/request_no_response")
    assert response == None


def test_request_long_response():
    request = main_pb2.TestRequest()
    request.data_ping = "Hello"

    response = main_pb2.TestResponseLong()
    call("test/request_long", request, response)

    for i in range(len(response.data_pong)):
        assert response.data_pong[i] == 'Hello...{}'.format(i)


def test_event():
    promise = Future()
    with events.pipe(rxop.take(1)).subscribe(lambda event: promise.set_result(event)):
        call("test/emit_event")

        event = promise.result()
        assert event[0] == "test/event"
        assert event[1].data == "testing event data"
